import logging
import copy
from datetime import datetime
from dateutil.relativedelta import relativedelta

from spaceone.core import cache, utils
from spaceone.core.manager import BaseManager
from spaceone.cost_analysis.error import *
from spaceone.cost_analysis.model.cost_model import Cost, MonthlyCost, CostQueryHistory
from spaceone.cost_analysis.manager.data_source_rule_manager import (
    DataSourceRuleManager,
)
from spaceone.cost_analysis.manager.identity_manager import IdentityManager

_LOGGER = logging.getLogger(__name__)


class CostManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cost_model: Cost = self.locator.get_model("Cost")
        self.monthly_cost_model: MonthlyCost = self.locator.get_model("MonthlyCost")
        self.cost_query_history_model: CostQueryHistory = self.locator.get_model(
            "CostQueryHistory"
        )
        self.data_source_rule_mgr: DataSourceRuleManager = self.locator.get_manager(
            "DataSourceRuleManager"
        )

    def create_cost(self, params: dict, execute_rollback=True):
        def _rollback(vo: Cost):
            _LOGGER.info(f"[create_cost._rollback] " f"Delete cost : {vo.cost_id} ")
            vo.delete()

        if "region_code" in params and "provider" in params:
            params["region_key"] = f'{params["provider"]}.{params["region_code"]}'

        billed_at = self._get_billed_at_from_billed_date(params["billed_date"])

        params["billed_year"] = billed_at.strftime("%Y")
        params["billed_month"] = billed_at.strftime("%Y-%m")

        params = self.data_source_rule_mgr.change_cost_data(params)

        cost_vo: Cost = self.cost_model.create(params)

        if execute_rollback:
            self.transaction.add_rollback(_rollback, cost_vo)

        return cost_vo

    def create_monthly_cost(self, params):
        return self.monthly_cost_model.create(params)

    def delete_cost(self, cost_id, domain_id):
        cost_vo: Cost = self.get_cost(cost_id, domain_id)
        self.delete_cost_by_vo(cost_vo)

    @staticmethod
    def delete_cost_by_vo(cost_vo: Cost):
        cost_vo.delete()

    def delete_cost_with_datasource(self, domain_id, data_source_id):
        _LOGGER.debug(f"[delete_cost_with_datasource] data_source_id: {data_source_id}")
        cost_vos = self.cost_model.filter(
            domain_id=domain_id, data_source_id=data_source_id
        )
        cost_vos.delete()

        monthly_cost_vos = self.monthly_cost_model.filter(
            domain_id=domain_id, data_source_id=data_source_id
        )
        monthly_cost_vos.delete()

        history_vos = self.cost_query_history_model.filter(
            domain_id=domain_id, data_source_id=data_source_id
        )
        history_vos.delete()

    def get_cost(
        self, cost_id, domain_id, workspace_id=None, user_projects: list = None
    ):
        conditions = {"cost_id": cost_id, "domain_id": domain_id}

        if workspace_id:
            conditions["workspace_id"] = workspace_id

        if user_projects:
            conditions["project_id"] = user_projects

        return self.cost_model.get(**conditions)

    def filter_costs(self, **conditions):
        return self.cost_model.filter(**conditions)

    def list_costs(self, query):
        query = self._change_filter_project_group_id(query)
        return self.cost_model.query(**query)

    def stat_costs(self, query):
        query = self._change_filter_project_group_id(query)
        return self.cost_model.stat(**query)

    def filter_monthly_costs(self, **conditions):
        return self.monthly_cost_model.filter(**conditions)

    def list_monthly_costs(self, query):
        query = self._change_filter_project_group_id(query)
        return self.monthly_cost_model.query(**query)

    def stat_monthly_costs(self, query):
        query = self._change_filter_project_group_id(query)
        return self.monthly_cost_model.stat(**query)

    def analyze_costs(self, query, target="SECONDARY_PREFERRED"):
        query["target"] = target
        query["date_field"] = "billed_date"
        query["date_field_format"] = "%Y-%m-%d"
        _LOGGER.debug(f"[analyze_costs] query: {query}")

        query = self._change_filter_project_group_id(query)
        return self.cost_model.analyze(**query)

    def analyze_monthly_costs(self, query, target="SECONDARY_PREFERRED"):
        query["target"] = target
        query["date_field"] = "billed_month"
        query["date_field_format"] = "%Y-%m"
        _LOGGER.debug(f"[analyze_monthly_costs] query: {query}")

        query = self._change_filter_project_group_id(query)
        return self.monthly_cost_model.analyze(**query)

    def analyze_yearly_costs(self, query, target="SECONDARY_PREFERRED"):
        query["target"] = target
        query["date_field"] = "billed_year"
        query["date_field_format"] = "%Y"
        _LOGGER.debug(f"[analyze_yearly_costs] query: {query}")

        query = self._change_filter_project_group_id(query)
        return self.monthly_cost_model.analyze(**query)

    @cache.cacheable(
        key="cost-analysis:stat-costs:monthly:{domain_id}:{data_source_id}:{query_hash}",
        expire=3600 * 24,
    )
    def stat_monthly_costs_with_cache(
        self, query, query_hash, domain_id, data_source_id
    ):
        return self.stat_monthly_costs(query)

    @cache.cacheable(
        key="cost-analysis:analyze-costs:daily:{domain_id}:{data_source_id}:{query_hash}",
        expire=3600 * 24,
    )
    def analyze_costs_with_cache(
        self, query, query_hash, domain_id, data_source_id, target="SECONDARY_PREFERRED"
    ):
        return self.analyze_costs(query, target)

    @cache.cacheable(
        key="cost-analysis:analyze-costs:monthly:{domain_id}:{data_source_id}:{query_hash}",
        expire=3600 * 24,
    )
    def analyze_monthly_costs_with_cache(
        self, query, query_hash, domain_id, data_source_id, target="SECONDARY_PREFERRED"
    ):
        return self.analyze_monthly_costs(query, target)

    @cache.cacheable(
        key="cost-analysis:analyze-costs:yearly:{domain_id}:{data_source_id}:{query_hash}",
        expire=3600 * 24,
    )
    def analyze_yearly_costs_with_cache(
        self, query, query_hash, domain_id, data_source_id, target="SECONDARY_PREFERRED"
    ):
        return self.analyze_yearly_costs(query, target)

    def analyze_costs_by_granularity(
        self, query: dict, domain_id: dict, data_source_id: dict
    ):
        self._check_date_range(query)
        granularity = query["granularity"]

        # Save query history to speed up data loading
        query_hash = utils.dict_to_hash(query)
        self.create_cost_query_history(query, query_hash, domain_id, data_source_id)

        if granularity == "DAILY":
            response = self.analyze_costs_with_cache(
                query, query_hash, domain_id, data_source_id
            )
        elif granularity == "MONTHLY":
            response = self.analyze_monthly_costs_with_cache(
                query, query_hash, domain_id, data_source_id
            )
        else:
            response = self.analyze_yearly_costs_with_cache(
                query, query_hash, domain_id, data_source_id
            )

        return response

    @cache.cacheable(
        key="cost-analysis:cost-query-history:{domain_id}:{data_source_id}:{query_hash}",
        expire=600,
    )
    def create_cost_query_history(self, query, query_hash, domain_id, data_source_id):
        def _rollback(history_vo):
            _LOGGER.info(
                f"[create_cost_query_history._rollback] Delete cost query history: {query_hash}"
            )
            history_vo.delete()

        history_model: CostQueryHistory = self.locator.get_model("CostQueryHistory")

        history_vos = history_model.filter(query_hash=query_hash, domain_id=domain_id)
        if history_vos.count() == 0:
            history_vo = history_model.create(
                {
                    "query_hash": query_hash,
                    "query_options": copy.deepcopy(query),
                    "data_source_id": data_source_id,
                    "domain_id": domain_id,
                }
            )

            self.transaction.add_rollback(_rollback, history_vo)
        else:
            history_vos[0].update({})

    def list_cost_query_history(self, query={}):
        history_model: CostQueryHistory = self.locator.get_model("CostQueryHistory")
        return history_model.query(**query)

    @staticmethod
    def remove_stat_cache(domain_id: str, data_source_id: str):
        cache.delete_pattern(
            f"cost-analysis:analyze-costs:*:{domain_id}:{data_source_id}:*"
        )
        cache.delete_pattern(
            f"cost-analysis:stat-costs:*:{domain_id}:{data_source_id}:*"
        )
        cache.delete_pattern(
            f"cost-analysis:cost-query-history:{domain_id}:{data_source_id}:*"
        )

    def _check_date_range(self, query):
        start_str = query.get("start")
        end_str = query.get("end")
        granularity = query.get("granularity")

        start = self._parse_start_time(start_str, granularity)
        end = self._parse_end_time(end_str, granularity)
        now = datetime.utcnow().date()

        if len(start_str) != len(end_str):
            raise ERROR_INVALID_DATE_RANGE(
                start=start_str,
                end=end_str,
                reason="Start date and end date must be the same format.",
            )

        if start >= end:
            raise ERROR_INVALID_DATE_RANGE(
                start=start_str,
                end=end_str,
                reason="End date must be greater than start date.",
            )

        if granularity == "DAILY":
            if start + relativedelta(months=1) < end:
                raise ERROR_INVALID_DATE_RANGE(
                    start=start_str,
                    end=end_str,
                    reason="Request up to a maximum of 1 month.",
                )

            if start + relativedelta(months=12) < now.replace(day=1):
                raise ERROR_INVALID_DATE_RANGE(
                    start=start_str,
                    end=end_str,
                    reason="For DAILY, you cannot request data older than 1 year.",
                )

        elif granularity == "MONTHLY":
            if start + relativedelta(months=12) < end:
                raise ERROR_INVALID_DATE_RANGE(
                    start=start_str,
                    end=end_str,
                    reason="Request up to a maximum of 12 months.",
                )

            if start + relativedelta(months=36) < now.replace(day=1):
                raise ERROR_INVALID_DATE_RANGE(
                    start=start_str,
                    end=end_str,
                    reason="For MONTHLY, you cannot request data older than 3 years.",
                )
        elif granularity == "YEARLY":
            if start + relativedelta(years=3) < now.replace(month=1, day=1):
                raise ERROR_INVALID_DATE_RANGE(
                    start=start_str,
                    end=end_str,
                    reason="For YEARLY, you cannot request data older than 3 years.",
                )

    def _parse_start_time(self, date_str, granularity):
        return self._convert_date_from_string(date_str.strip(), "start", granularity)

    def _parse_end_time(self, date_str, granularity):
        end = self._convert_date_from_string(date_str.strip(), "end", granularity)

        if granularity == "YEARLY":
            return end + relativedelta(years=1)
        elif granularity == "MONTHLY":
            return end + relativedelta(months=1)
        else:
            return end + relativedelta(days=1)

    @staticmethod
    def _convert_date_from_string(date_str, key, granularity):
        if granularity == "YEARLY":
            date_format = "%Y"
            date_type = "YYYY"
        elif granularity == "MONTHLY":
            if len(date_str) == 4:
                date_format = "%Y"
                date_type = "YYYY"
            else:
                date_format = "%Y-%m"
                date_type = "YYYY-MM"
        else:
            if len(date_str) == 4:
                date_format = "%Y"
                date_type = "YYYY"
            elif len(date_str) == 7:
                date_format = "%Y-%m"
                date_type = "YYYY-MM"
            else:
                date_format = "%Y-%m-%d"
                date_type = "YYYY-MM-DD"

        try:
            return datetime.strptime(date_str, date_format).date()
        except Exception as e:
            raise ERROR_INVALID_PARAMETER_TYPE(key=key, type=date_type)

    @staticmethod
    def _get_billed_at_from_billed_date(billed_date):
        date_format = "%Y-%m-%d"

        try:
            return datetime.strptime(billed_date, date_format)
        except Exception as e:
            raise ERROR_INVALID_PARAMETER_TYPE(key="billed_date", type="YYYY-MM-DD")

    def _change_filter_project_group_id(self, query: dict) -> dict:
        change_filter = []
        self.identity_mgr = None

        for condition in query.get("filter", []):
            key = condition.get("k", condition.get("key"))
            value = condition.get("v", condition.get("value"))
            operator = condition.get("o", condition.get("operator"))

            if key == "project_group_id":
                if self.identity_mgr is None:
                    self.identity_mgr: IdentityManager = self.locator.get_manager(
                        "IdentityManager"
                    )

                project_groups_info = self.identity_mgr.list_project_groups(
                    {
                        "query": {
                            "only": ["project_group_id"],
                            "filter": [{"k": key, "v": value, "o": operator}],
                        }
                    }
                )

                project_group_ids = [
                    project_group_info["project_group_id"]
                    for project_group_info in project_groups_info.get("results", [])
                ]

                project_ids = []

                for project_group_id in project_group_ids:
                    projects_info = self.identity_mgr.get_projects_in_project_group(
                        project_group_id
                    )
                    project_ids.extend(
                        [
                            project_info["project_id"]
                            for project_info in projects_info.get("results", [])
                        ]
                    )

                project_ids = list(set(project_ids))
                change_filter.append({"k": "project_id", "v": project_ids, "o": "in"})

            else:
                change_filter.append(condition)

        query["filter"] = change_filter
        return query



from spaceone.api.cost_analysis.v1 import cost_query_set_pb2, cost_query_set_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class CostQuerySet(BaseAPI, cost_query_set_pb2_grpc.CostQuerySetServicer):

    pb2 = cost_query_set_pb2
    pb2_grpc = cost_query_set_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostQuerySetService', metadata) as cost_query_set_service:
            return self.locator.get_info('CostQuerySetInfo', cost_query_set_service.create(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostQuerySetService', metadata) as cost_query_set_service:
            return self.locator.get_info('CostQuerySetInfo', cost_query_set_service.update(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostQuerySetService', metadata) as cost_query_set_service:
            cost_query_set_service.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostQuerySetService', metadata) as cost_query_set_service:
            return self.locator.get_info('CostQuerySetInfo', cost_query_set_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostQuerySetService', metadata) as cost_query_set_service:
            cost_query_set_vos, total_count = cost_query_set_service.list(params)
            return self.locator.get_info('CostQuerySetsInfo',
                                         cost_query_set_vos,
                                         total_count,
                                         minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostQuerySetService', metadata) as cost_query_set_service:
            return self.locator.get_info('StatisticsInfo', cost_query_set_service.stat(params))



import logging

from spaceone.core.manager import BaseManager
from spaceone.cost_analysis.model.data_source_model import DataSource

_LOGGER = logging.getLogger(__name__)


class DataSourceManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_source_model: DataSource = self.locator.get_model("DataSource")

    def register_data_source(self, params):
        def _rollback(data_source_vo):
            _LOGGER.info(
                f"[register_data_source._rollback] "
                f"Delete data source : {data_source_vo.name} "
                f"({data_source_vo.data_source_id})"
            )
            data_source_vo.delete()

        data_source_vo: DataSource = self.data_source_model.create(params)
        self.transaction.add_rollback(_rollback, data_source_vo)

        return data_source_vo

    def update_data_source(self, params):
        data_source_vo: DataSource = self.get_data_source(
            params["data_source_id"], params["domain_id"]
        )
        return self.update_data_source_by_vo(params, data_source_vo)

    def update_data_source_by_vo(self, params, data_source_vo):
        def _rollback(old_data):
            _LOGGER.info(
                f"[update_data_source_by_vo._rollback] Revert Data : "
                f'{old_data["data_source_id"]}'
            )
            data_source_vo.update(old_data)

        self.transaction.add_rollback(_rollback, data_source_vo.to_dict())
        return data_source_vo.update(params)

    def deregister_data_source(self, data_source_id, domain_id):
        data_source_vo: DataSource = self.get_data_source(data_source_id, domain_id)
        data_source_vo.delete()

    @staticmethod
    def deregister_data_source_by_vo(data_source_vo):
        data_source_vo.delete()

    def get_data_source(
        self, data_source_id: str, domain_id: str, workspace_id: str = None
    ):
        conditions = {"data_source_id": data_source_id, "domain_id": domain_id}

        if workspace_id:
            conditions["workspace_id"] = workspace_id
        return self.data_source_model.get(**conditions)

    def filter_data_sources(self, **conditions):
        return self.data_source_model.filter(**conditions)

    def list_data_sources(self, query={}):
        return self.data_source_model.query(**query)

    def stat_data_sources(self, query):
        return self.data_source_model.stat(**query)



from spaceone.cost_analysis.info.common_info import *
from spaceone.cost_analysis.info.data_source_info import *
from spaceone.cost_analysis.info.data_source_rule_info import *
from spaceone.cost_analysis.info.cost_info import *
from spaceone.cost_analysis.info.budget_info import *
from spaceone.cost_analysis.info.budget_usage_info import *
from spaceone.cost_analysis.info.cost_query_set_info import *
from spaceone.cost_analysis.info.job_info import *
from spaceone.cost_analysis.info.job_task_info import *



import logging
from typing import Union
from spaceone.core.service import BaseService, transaction, convert_model
from spaceone.cost_analysis.plugin.data_source.model import (JobGetTaskRequest, TasksResponse)

_LOGGER = logging.getLogger(__name__)


class JobService(BaseService):

    @transaction
    @convert_model
    def get_tasks(self, params: JobGetTaskRequest) -> Union[TasksResponse, dict]:
        """ Get job tasks

        Args:
            params (JobGetTaskRequest): {
                'options': 'dict',                  # Required
                'secret_data': 'dict',              # Required
                'schema': 'str',
                'start': 'str',
                'last_synchronized_at': 'datetime',
                'domain_id': 'str'                  # Required
            }

        Returns:
            TasksResponse: {
                'tasks': 'list',
                'changed': 'list'
            }

        """

        func = self.get_plugin_method('get_tasks')
        response = func(params.dict())
        return TasksResponse(**response)



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class JobTask(MongoModel):
    job_task_id = StringField(max_length=40, generate_id="job-task", unique=True)
    status = StringField(
        max_length=20,
        default="PENDING",
        choices=("PENDING", "IN_PROGRESS", "SUCCESS", "FAILURE", "CANCELED"),
    )
    options = DictField()
    created_count = IntField(default=0)
    error_code = StringField(max_length=254, default=None, null=True)
    error_message = StringField(default=None, null=True)
    resource_group = StringField(max_length=40, choices=["DOMAIN", "WORKSPACE"])
    job_id = StringField(max_length=40, required=True)
    data_source_id = StringField(max_length=40, required=True)
    workspace_id = StringField(max_length=40, default=None, null=True)
    domain_id = StringField(max_length=40, required=True)
    created_at = DateTimeField(auto_now_add=True)
    started_at = DateTimeField(default=None, null=True)
    updated_at = DateTimeField(auto_now=True)
    finished_at = DateTimeField(default=None, null=True)

    meta = {
        "updatable_fields": [
            "status",
            "created_count",
            "error_code",
            "error_message",
            "started_at",
            "updated_at",
            "finished_at",
        ],
        "minimal_fields": [
            "job_task_id",
            "status",
            "created_count",
            "job_id",
            "data_source_id",
            "workspace_id",
        ],
        "ordering": ["-created_at"],
        "indexes": [
            # 'job_task_id',
            "status",
            "job_id",
            "resource_group",
            "data_source_id",
            "workspace_id",
            "domain_id",
            "created_at",
        ],
    }



from spaceone.core.error import *


class ERROR_INVALID_DATE_RANGE(ERROR_INVALID_ARGUMENT):
    _message = "{reason}"


class ERROR_NOT_SUPPORT_QUERY_OPTION(ERROR_INVALID_ARGUMENT):
    _message = "Not support query option. (query_option = {query_option})"



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class PluginInfo(EmbeddedDocument):
    plugin_id = StringField(max_length=40)
    version = StringField(max_length=255)
    options = DictField(default={})
    metadata = DictField(default={})
    secret_id = StringField(max_length=40, default=None, null=True)
    schema_id = StringField(max_length=255, default=None, null=True)
    upgrade_mode = StringField(
        max_length=255, choices=("AUTO", "MANUAL"), default="AUTO"
    )

    def to_dict(self):
        return dict(self.to_mongo())


class SecretFilter(EmbeddedDocument):
    state = StringField(
        max_length=20, default="ENABLED", choices=("ENABLED", "DISABLED")
    )
    secrets = ListField(StringField(max_length=40), defualt=None, null=True)
    service_accounts = ListField(StringField(max_length=40), default=None, null=True)
    schemas = ListField(StringField(max_length=40), default=None, null=True)

    def to_dict(self):
        return dict(self.to_mongo())


class DataSource(MongoModel):
    data_source_id = StringField(max_length=40, generate_id="ds", unique=True)
    name = StringField(max_length=255, unique_with="domain_id")
    state = StringField(
        max_length=20, default="ENABLED", choices=("ENABLED", "DISABLED")
    )
    data_source_type = StringField(max_length=20, choices=("LOCAL", "EXTERNAL"))
    secret_type = StringField(
        max_length=32,
        default="MANUAL",
        choices=("MANUAL", "USE_SERVICE_ACCOUNT_SECRET"),
    )
    secret_filter = EmbeddedDocumentField(SecretFilter, default=None, null=True)
    provider = StringField(max_length=40, default=None, null=True)
    plugin_info = EmbeddedDocumentField(PluginInfo, default=None, null=True)
    template = DictField(default={})
    tags = DictField(default={})
    cost_tag_keys = ListField(StringField())
    cost_additional_info_keys = ListField(StringField())
    cost_data_keys = ListField(StringField())
    resource_group = StringField(
        max_length=255, default=None, null=True, choices=("DOMAIN", "WORKSPACE")
    )
    workspace_id = StringField(max_length=40)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    last_synchronized_at = DateTimeField(default=None, null=True)

    meta = {
        "updatable_fields": [
            "name",
            "state",
            "plugin_info",
            "secret_filter",
            "template",
            "tags",
            "last_synchronized_at",
            "cost_tag_keys",
            "cost_additional_info_keys",
            "cost_data_keys",
        ],
        "minimal_fields": [
            "data_source_id",
            "workspace_id",
            "name",
            "state",
            "data_source_type",
            "secret_type",
            "provider",
        ],
        "ordering": ["name"],
        "indexes": [
            "state",
            "data_source_type",
            "provider",
            "resource_group",
            "workspace_id",
            "domain_id"
        ],
    }



LOG = {
    'filters': {
        'masking': {
            'rules': {
                'DataSource.verify': [
                    'secret_data'
                ],
                'Job.get_tasks': [
                    'secret_data'
                ],
                'Cost.get_data': [
                    'secret_data'
                ]
            }
        }
    }
}



import logging

from spaceone.core.manager import BaseManager
from spaceone.cost_analysis.model.budget_model import Budget

_LOGGER = logging.getLogger(__name__)


class BudgetManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.budget_model: Budget = self.locator.get_model("Budget")

    def create_budget(self, params: dict) -> Budget:
        def _rollback(vo: Budget):
            _LOGGER.info(
                f"[create_budget._rollback] "
                f"Delete budget : {vo.name} "
                f"({vo.budget_id})"
            )
            vo.delete()

        budget_vo: Budget = self.budget_model.create(params)
        self.transaction.add_rollback(_rollback, budget_vo)

        return budget_vo

    def update_budget_by_vo(self, params, budget_vo):
        def _rollback(old_data):
            _LOGGER.info(
                f"[update_budget_by_vo._rollback] Revert Data : "
                f'{old_data["budget_id"]}'
            )
            budget_vo.update(old_data)

        self.transaction.add_rollback(_rollback, budget_vo.to_dict())
        return budget_vo.update(params)

    @staticmethod
    def delete_budget_by_vo(budget_vo: Budget) -> None:
        budget_vo.delete()

    def get_budget(
        self, budget_id: str, domain_id: str, workspace_id: str = None, project_id=None
    ) -> Budget:
        conditions = {"budget_id": budget_id, "domain_id": domain_id}

        if workspace_id:
            conditions["workspace_id"] = workspace_id

        if project_id:
            conditions["project_id"] = project_id

        return self.budget_model.get(**conditions)

    def filter_budgets(self, **conditions):
        return self.budget_model.filter(**conditions)

    def list_budgets(self, query: dict):
        return self.budget_model.query(**query)

    def stat_budgets(self, query):
        return self.budget_model.stat(**query)



import logging
from datetime import datetime
from dateutil.rrule import rrule, MONTHLY, YEARLY

from spaceone.core.service import *
from spaceone.cost_analysis.error import *
from spaceone.cost_analysis.manager.data_source_manager import DataSourceManager
from spaceone.cost_analysis.manager.budget_manager import BudgetManager
from spaceone.cost_analysis.manager.budget_usage_manager import BudgetUsageManager
from spaceone.cost_analysis.manager.identity_manager import IdentityManager
from spaceone.cost_analysis.model.budget_model import Budget
from spaceone.cost_analysis.model.data_source_model import DataSource
from spaceone.cost_analysis.model.budget_usage_model import BudgetUsage

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class BudgetService(BaseService):
    resource = "Budget"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.budget_mgr: BudgetManager = self.locator.get_manager("BudgetManager")

    @transaction(
        permission="cost-analysis:Budget.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(
        [
            "data_source_id",
            "time_unit",
            "start",
            "end",
            "resource_group",
            "domain_id",
        ]
    )
    def create(self, params):
        """Register budget

        Args:
            params (dict): {
                'data_source_id': 'str',    # required
                'name': 'str',
                'limit': 'float',
                'planned_limits': 'list',
                'time_unit': 'str',         # required
                'start': 'str',             # required
                'end': 'str',               # required
                'provider_filter': 'dict',
                'notifications': 'list',
                'tags': 'dict',
                'resource_group': 'str',    # required
                'project_id': 'str',
                'workspace_id': 'str',
                'domain_id': 'str'          # injected from auth
            }

        Returns:
            budget_vo (object)
        """

        domain_id = params["domain_id"]
        workspace_id = params["workspace_id"]
        data_source_id = params["data_source_id"]
        project_id = params.get("project_id")
        limit = params.get("limit")
        planned_limits = params.get("planned_limits", [])
        time_unit = params["time_unit"]
        start = params["start"]
        end = params["end"]
        provider_filter = params.get("provider_filter", {})
        provider_filter_state = provider_filter.get("state", "DISABLED")
        notifications = params.get("notifications", [])
        resource_group = params["resource_group"]
        self._check_time_period(start, end)

        identity_mgr: IdentityManager = self.locator.get_manager("IdentityManager")
        if resource_group == "WORKSPACE":
            identity_mgr.check_workspace(workspace_id, domain_id)
            params["project_id"] = "*"
            project_id = "*"
        else:
            identity_mgr.get_project(project_id, domain_id)

        # Check Provider Filter
        if provider_filter_state == "ENABLED":
            if len(provider_filter.get("providers", [])) == 0:
                raise ERROR_PROVIDER_FILTER_IS_EMPTY()
        else:
            params["provider_filter"] = {"state": "DISABLED", "providers": []}

        data_source_mgr: DataSourceManager = self.locator.get_manager(
            "DataSourceManager"
        )

        data_source_vos = data_source_mgr.filter_data_sources(
            data_source_id=data_source_id,
            workspace_id=[workspace_id, "*"],
            domain_id=domain_id,
        )

        data_source_vo = data_source_vos.first()

        data_source_metadata = data_source_vo.plugin_info.metadata
        params["currency"] = data_source_metadata.get("currency", "USD")

        if time_unit == "TOTAL":
            if limit is None:
                raise ERROR_REQUIRED_PARAMETER(key="limit")

            params["planned_limits"] = None

        else:
            # Check Planned Limits
            self._check_planned_limits(start, end, time_unit, planned_limits)

            params["limit"] = 0
            for planned_limit in planned_limits:
                params["limit"] += planned_limit.get("limit", 0)

        # Check Notifications
        self._check_notifications(notifications, project_id)

        # Check Duplicated Budget
        budget_vos = self.budget_mgr.filter_budgets(
            data_source_id=data_source_id,
            project_id=project_id,
            workspace_id=workspace_id,
            domain_id=domain_id,
        )
        if budget_vos.count() > 0:
            raise ERROR_BUDGET_ALREADY_EXIST(
                data_source_id=data_source_id,
                workspace_id=workspace_id,
                target=project_id,
            )

        budget_vo = self.budget_mgr.create_budget(params)

        # Create budget usages
        budget_usage_mgr: BudgetUsageManager = self.locator.get_manager(
            "BudgetUsageManager"
        )
        budget_usage_mgr.create_budget_usages(budget_vo)
        budget_usage_mgr.update_cost_usage(budget_vo)
        budget_usage_mgr.notify_budget_usage(budget_vo)

        return budget_vo

    @transaction(
        permission="cost-analysis:Budget.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(["budget_id", "domain_id"])
    # @change_date_value(['end'])
    def update(self, params):
        """Update budget

        Args:
            params (dict): {
                'budget_id': 'str',         # required
                'name': 'str',
                'limit': 'float',
                'planned_limits': 'list',
                'tags': 'dict'
                'workspace_id', 'str',      # injected from auth (optional)
                'domain_id': 'str'          # injected from auth
            }

        Returns:
            budget_vo (object)
        """

        budget_id = params["budget_id"]
        workspace_id = params.get("workspace_id")
        domain_id = params["domain_id"]
        planned_limits = params.get("planned_limits")

        budget_usage_mgr: BudgetUsageManager = self.locator.get_manager(
            "BudgetUsageManager"
        )

        budget_vo: Budget = self.budget_mgr.get_budget(
            budget_id, domain_id, workspace_id
        )

        # Check limit and Planned Limits
        budget_vo = self.budget_mgr.update_budget_by_vo(params, budget_vo)

        if "name" in params:
            budget_usage_vos = budget_usage_mgr.filter_budget_usages(
                budget_id=budget_id,
                domain_id=domain_id,
            )
            for budget_usage_vo in budget_usage_vos:
                budget_usage_mgr.update_budget_usage_by_vo(
                    {"name": params["name"]}, budget_usage_vo
                )
        # Check DataSource exists
        data_source_mgr = self.locator.get_manager("DataSourceManager")
        data_source_mgr.get_data_source(budget_vo.data_source_id, domain_id)

        budget_usage_mgr.update_cost_usage(budget_vo)
        budget_usage_mgr.notify_budget_usage(budget_vo)

        return budget_vo

    @transaction(
        permission="cost-analysis:Budget.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(["budget_id", "workspace_id", "domain_id"])
    def set_notification(self, params):
        """Set budget notification

        Args:
            params (dict): {
                'budget_id': 'str',
                'notifications': 'list',
                'workspace_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            budget_vo (object)
        """
        budget_id = params["budget_id"]
        workspace_id = params.get("workspace_id")
        domain_id = params["domain_id"]
        notifications = params.get("notifications", [])

        budget_vo: Budget = self.budget_mgr.get_budget(
            budget_id, domain_id, workspace_id
        )

        # Check Notifications
        self._check_notifications(notifications, budget_vo.project_id)
        params["notifications"] = notifications

        return self.budget_mgr.update_budget_by_vo(params, budget_vo)

    @transaction(
        permission="cost-analysis:Budget.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(["budget_id", "domain_id"])
    def delete(self, params):
        """Deregister budget

        Args:
            params (dict): {
                'budget_id': 'str',     # required
                'workspace_id': 'str',  # injected from auth (optional)
                'domain_id': 'str'      # injected from auth
            }

        Returns:
            None
        """

        budget_vo: Budget = self.budget_mgr.get_budget(
            params["budget_id"], params["domain_id"], params.get("workspace_id")
        )
        self.budget_mgr.delete_budget_by_vo(budget_vo)

    @transaction(
        permission="cost-analysis:Budget.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @check_required(["budget_id", "domain_id"])
    def get(self, params):
        """Get budget

        Args:
            params (dict): {
                'budget_id': 'str',         # required
                'user_projects': 'list',    # injected from auth (optional)
                'workspace_id': 'str',      # injected from auth (optional)
                'domain_id': 'str',         # injected from auth
            }

        Returns:
            budget_vo (object)
        """

        budget_id = params["budget_id"]
        domain_id = params["domain_id"]
        workspace_id = params.get("workspace_id")
        project_id = params.get("user_projects")

        return self.budget_mgr.get_budget(
            budget_id, domain_id, workspace_id, project_id
        )

    @transaction(
        permission="cost-analysis:Budget.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @check_required(["domain_id"])
    @append_query_filter(
        [
            "budget_id",
            "name",
            "time_unit",
            "data_source_id",
            "project_id",
            "user_projects",
            "workspace_id",
            "domain_id",
        ]
    )
    @append_keyword_filter(["budget_id", "name"])
    def list(self, params):
        """List budgets

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)',
                'budget_id': 'str',
                'name': 'str',
                'time_unit': 'str',
                'data_source_id': 'str',
                'project_id': 'str',
                'user_projects': 'list',
                'workspace_id': 'str',
                'domain_id': 'str',
            }

        Returns:
            budget_vos (object)
            total_count
        """

        query: dict = self._set_user_project_or_project_group_filter(params)
        return self.budget_mgr.list_budgets(query)

    @transaction(
        permission="cost-analysis:Budget.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @check_required(["query", "domain_id"])
    @append_query_filter(["user_projects", "workspace_id", "domain_id"])
    @append_keyword_filter(["budget_id", "name"])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta,
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get("query", {})
        return self.budget_mgr.stat_budgets(query)

    @staticmethod
    def _check_time_period(start, end):
        if start >= end:
            raise ERROR_INVALID_TIME_RANGE(start=start, end=end)

    def _check_planned_limits(self, start, end, time_unit, planned_limits):
        planned_limits_dict = self._convert_planned_limits_data_type(planned_limits)
        date_format = "%Y-%m"

        try:
            start_dt = datetime.strptime(start, date_format)
        except Exception as e:
            raise ERROR_INVALID_PARAMETER_TYPE(key="start", type=date_format)

        try:
            end_dt = datetime.strptime(end, date_format)
        except Exception as e:
            raise ERROR_INVALID_PARAMETER_TYPE(key="end", type=date_format)

        for dt in rrule(MONTHLY, dtstart=start_dt, until=end_dt):
            date_str = dt.strftime(date_format)
            if date_str not in planned_limits_dict:
                raise ERROR_NO_DATE_IN_PLANNED_LIMITS(date=date_str)

            del planned_limits_dict[date_str]

        if len(planned_limits_dict.keys()) > 0:
            raise ERROR_DATE_IS_WRONG(date=list(planned_limits_dict.keys()))

    @staticmethod
    def _convert_planned_limits_data_type(planned_limits):
        planned_limits_dict = {}

        for planned_limit in planned_limits:
            date = planned_limit.get("date")
            limit = planned_limit.get("limit", 0)
            if date is None:
                raise ERROR_DATE_IS_REQUIRED(value=planned_limit)

            if limit < 0:
                raise ERROR_LIMIT_IS_WRONG(value=planned_limit)

            planned_limits_dict[date] = limit

        return planned_limits_dict

    @staticmethod
    def _check_notifications(notifications, project_id):
        if len(notifications) > 0 and project_id is None:
            raise ERROR_NOTIFICATION_IS_NOT_SUPPORTED_IN_PROJECT(target=project_id)

        for notification in notifications:
            unit = notification.get("unit")
            notification_type = notification.get("notification_type")
            threshold = notification.get("threshold", 0)

            if unit not in ["PERCENT", "ACTUAL_COST"]:
                raise ERROR_UNIT_IS_REQUIRED(value=notification)

            if notification_type not in ["CRITICAL", "WARNING"]:
                raise ERROR_NOTIFICATION_TYPE_IS_REQUIRED(value=notification)

            if threshold < 0:
                raise ERROR_THRESHOLD_IS_WRONG(value=notification)

            if unit == "PERCENT":
                if threshold > 100:
                    raise ERROR_THRESHOLD_IS_WRONG_IN_PERCENT_TYPE(value=notification)

    @staticmethod
    def _set_user_project_or_project_group_filter(params):
        query = params.get("query", {})
        query["filter"] = query.get("filter", [])

        if "user_projects" in params:
            user_projects = params["user_projects"] + [None]
            query["filter"].append(
                {"k": "user_projects", "v": user_projects, "o": "in"}
            )

        return query



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class CostQuerySet(MongoModel):
    cost_query_set_id = StringField(max_length=40, generate_id="query", unique=True)
    name = StringField(max_length=255)
    options = DictField(default={})
    tags = DictField(default={})
    user_id = StringField(max_length=40)
    data_source_id = StringField(max_length=40)
    workspace_id = StringField(max_length=40, default=None, null=True)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        "updatable_fields": ["name", "options", "tags"],
        "minimal_fields": [
            "cost_query_set_id",
            "name",
            "user_id",
            "data_source_id",
            "workspace_id",
        ],
        "ordering": ["name"],
        "indexes": ["name", "user_id", "data_source_id", "workspace_id", "domain_id"],
    }



import functools
from spaceone.api.cost_analysis.v1 import cost_query_set_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.cost_analysis.model.cost_query_set_model import CostQuerySet

__all__ = ["CostQuerySetInfo", "CostQuerySetsInfo"]


def CostQuerySetInfo(cost_query_set_vo: CostQuerySet, minimal=False):
    info = {
        "cost_query_set_id": cost_query_set_vo.cost_query_set_id,
        "name": cost_query_set_vo.name,
        "user_id": cost_query_set_vo.user_id,
        "workspace_id": cost_query_set_vo.workspace_id,
    }

    if not minimal:
        info.update(
            {
                "options": change_struct_type(cost_query_set_vo.options),
                "tags": change_struct_type(cost_query_set_vo.tags),
                "data_source_id": cost_query_set_vo.data_source_id,
                "domain_id": cost_query_set_vo.domain_id,
                "created_at": utils.datetime_to_iso8601(cost_query_set_vo.created_at),
                "updated_at": utils.datetime_to_iso8601(cost_query_set_vo.updated_at),
            }
        )

    return cost_query_set_pb2.CostQuerySetInfo(**info)


def CostQuerySetsInfo(cost_query_set_vos, total_count, **kwargs):
    return cost_query_set_pb2.CostQuerySetsInfo(
        results=list(
            map(functools.partial(CostQuerySetInfo, **kwargs), cost_query_set_vos)
        ),
        total_count=total_count,
    )



import logging

from spaceone.core import config
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.core import utils
from spaceone.core.auth.jwt.jwt_util import JWTUtil

_LOGGER = logging.getLogger(__name__)


class SecretManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        token = self.transaction.get_meta("token")
        self.token_type = JWTUtil.get_value_from_token(token, "typ")
        self.secret_connector: SpaceConnector = self.locator.get_connector(
            "SpaceConnector", service="secret"
        )

    def create_secret(
        self,
        secret_data: dict,
        resource_group: str,
        schema_id: str,
        workspace_id: str,
    ):
        def _rollback(secret_id: str):
            _LOGGER.info(f"[create_secret._rollback] Delete secret : {secret_id}")
            self.delete_secret(secret_id)

        params = {
            "name": utils.generate_id("secret-cost-data-source"),
            "resource_group": resource_group,
            "data": secret_data,
            "schema_id": schema_id,
            "workspace_id": workspace_id,
        }

        response = self.secret_connector.dispatch("Secret.create", params)

        _LOGGER.debug(f"[_create_secret] {response}")
        secret_id = response["secret_id"]

        self.transaction.add_rollback(_rollback, secret_id)

        return secret_id

    def delete_secret(self, secret_id: str):
        self.secret_connector.dispatch("Secret.delete", {"secret_id": secret_id})

    def list_secrets(self, query: dict, domain_id: str = None) -> dict:
        params = {"query": query}

        if self.token_type == "SYSTEM_TOKEN":
            return self.secret_connector.dispatch(
                "Secret.list", params, x_domain_id=domain_id
            )
        else:
            return self.secret_connector.dispatch("Secret.list", params)

    def get_secret(self, secret_id: str):
        return self.secret_connector.dispatch("Secret.get", {"secret_id": secret_id})

    def get_secret_data(self, secret_id, domain_id):
        system_token = config.get_global("TOKEN")

        response = self.secret_connector.dispatch(
            "Secret.get_data",
            {"secret_id": secret_id, "domain_id": domain_id},
            token=system_token,
        )
        return response["data"]



from spaceone.api.cost_analysis.v1 import job_task_pb2, job_task_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class JobTask(BaseAPI, job_task_pb2_grpc.JobTaskServicer):
    pb2 = job_task_pb2
    pb2_grpc = job_task_pb2_grpc

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service("JobTaskService", metadata) as job_task_service:
            return self.locator.get_info("JobTaskInfo", job_task_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service("JobTaskService", metadata) as job_task_service:
            job_task_vos, total_count = job_task_service.list(params)
            return self.locator.get_info(
                "JobTasksInfo",
                job_task_vos,
                total_count,
                minimal=self.get_minimal(params),
            )

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service("JobTaskService", metadata) as job_task_service:
            return self.locator.get_info(
                "StatisticsInfo", job_task_service.stat(params)
            )



from spaceone.core.pygrpc import BaseAPI
from spaceone.api.cost_analysis.plugin import job_pb2, job_pb2_grpc
from spaceone.cost_analysis.plugin.data_source.service.job_service import JobService


class Job(BaseAPI, job_pb2_grpc.JobServicer):

    pb2 = job_pb2
    pb2_grpc = job_pb2_grpc

    def get_tasks(self, request, context):
        params, metadata = self.parse_request(request, context)
        job_svc = JobService(metadata)
        response: dict = job_svc.get_tasks(params)
        return self.dict_to_message(response)



import logging
import copy
from typing import List, Union
from datetime import datetime, timedelta

from spaceone.core.error import *
from spaceone.core import queue, utils, config
from spaceone.core.manager import BaseManager
from spaceone.cost_analysis.model.job_model import Job
from spaceone.cost_analysis.model.cost_model import CostQueryHistory
from spaceone.cost_analysis.manager.cost_manager import CostManager

_LOGGER = logging.getLogger(__name__)


class JobManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.job_model: Job = self.locator.get_model("Job")
        self.job_timeout = config.get_global("JOB_TIMEOUT", 7200)
        self.cost_mgr: CostManager = self.locator.get_manager("CostManager")

    def is_job_running(self, data_source_id, domain_id):
        job_vos: List[Job] = self.job_model.filter(
            data_source_id=data_source_id, domain_id=domain_id, status="IN_PROGRESS"
        )

        running_job_count = job_vos.count()

        for job_vo in job_vos:
            if datetime.utcnow() > (
                job_vo.created_at + timedelta(seconds=self.job_timeout)
            ):
                self.change_timeout_status(job_vo)
                running_job_count -= 1

        return running_job_count > 0

    def create_job(
        self,
        resource_group: str,
        data_source_id: str,
        workspace_id: str,
        domain_id: str,
        job_options: dict,
        total_tasks: int,
        changed: Union[list, None] = None,
    ):
        job_options["no_preload_cache"] = job_options.get("no_preload_cache", False)
        job_options["start"] = job_options.get("start")
        job_options["sync_mode"] = job_options.get("sync_mode", "SCHEDULED")

        data = {
            "resource_group": resource_group,
            "data_source_id": data_source_id,
            "workspace_id": workspace_id,
            "domain_id": domain_id,
            "options": job_options,
        }

        if total_tasks:
            data.update({"total_tasks": total_tasks, "remained_tasks": total_tasks})

        if changed:
            data["changed"] = changed

        job_vo = self.job_model.create(data)

        _LOGGER.debug(f"[create_job] create job: {job_vo.job_id}")
        return job_vo

    def update_job_by_vo(self, params, job_vo):
        return job_vo.update(params)

    def get_job(
        self, job_id: str, domain_id: str, workspace_id: Union[str, list] = None
    ):
        conditions = {
            "job_id": job_id,
            "domain_id": domain_id,
        }

        if workspace_id:
            conditions["workspace_id"] = workspace_id
        return self.job_model.get(**conditions)

    def filter_jobs(self, **conditions):
        return self.job_model.filter(**conditions)

    def list_jobs(self, query={}):
        return self.job_model.query(**query)

    def stat_jobs(self, query):
        return self.job_model.stat(**query)

    def preload_cost_stat_queries(self, domain_id, data_source_id):
        cost_query_cache_time = config.get_global("COST_QUERY_CACHE_TIME", 4)
        cache_time = datetime.utcnow() - timedelta(days=cost_query_cache_time)

        query = {
            "filter": [
                {"k": "domain_id", "v": domain_id, "o": "eq"},
                {"k": "data_source_id", "v": data_source_id, "o": "eq"},
                {"k": "updated_at", "v": cache_time, "o": "gte"},
            ]
        }

        _LOGGER.debug(
            f"[_preload_cost_stat_queries] cost_query_cache_time: {cost_query_cache_time} days"
        )

        history_vos, total_count = self.cost_mgr.list_cost_query_history(query)
        for history_vo in history_vos:
            _LOGGER.debug(
                f"[_preload_cost_stat_queries] create query cache: {history_vo.query_hash}"
            )
            self._create_cache_by_history(history_vo, domain_id)

    def _create_cache_by_history(self, history_vo: CostQueryHistory, domain_id):
        query = history_vo.query_options
        query_hash = history_vo.query_hash
        data_source_id = history_vo.data_source_id

        # Original Date Range
        self._create_cache(copy.deepcopy(query), query_hash, domain_id, data_source_id)

    def _create_cache(self, query, query_hash, domain_id, data_source_id):
        if granularity := query.get("granularity"):
            if granularity == "DAILY":
                self.cost_mgr.analyze_costs_with_cache(
                    query, query_hash, domain_id, data_source_id
                )
            elif granularity == "MONTHLY":
                self.cost_mgr.analyze_monthly_costs_with_cache(
                    query, query_hash, domain_id, data_source_id
                )
            elif granularity == "YEARLY":
                self.cost_mgr.analyze_yearly_costs_with_cache(
                    query, query_hash, domain_id, data_source_id
                )
        else:
            self.cost_mgr.stat_monthly_costs_with_cache(
                query, query_hash, domain_id, data_source_id
            )

    @staticmethod
    def decrease_remained_tasks(job_vo: Job):
        return job_vo.decrement("remained_tasks", 1)

    @staticmethod
    def change_success_status(job_vo: Job):
        _LOGGER.info(f"[change_success_status] job success: {job_vo.job_id}")

        return job_vo.update({"status": "SUCCESS", "finished_at": datetime.utcnow()})

    @staticmethod
    def change_canceled_status(job_vo: Job):
        _LOGGER.error(f"[change_canceled_status], job canceled ({job_vo.job_id})")

        return job_vo.update({"status": "CANCELED", "finished_at": datetime.utcnow()})

    @staticmethod
    def change_timeout_status(job_vo: Job):
        _LOGGER.error(f"[change_timeout_status] job timeout: {job_vo.job_id}")

        return job_vo.update({"status": "TIMEOUT", "finished_at": datetime.utcnow()})

    @staticmethod
    def change_error_status(job_vo: Job, e):
        if not isinstance(e, ERROR_BASE):
            e = ERROR_UNKNOWN(message=str(e))

        _LOGGER.error(
            f"[change_error_status] job error ({job_vo.job_id}): {e.message}",
            exc_info=True,
        )

        job_vo.update(
            {
                "status": "FAILURE",
                "error_code": e.error_code,
                "error_message": e.message,
                "finished_at": datetime.utcnow(),
            }
        )



from typing import Union
from pydantic import BaseModel, Field

__all__ = ['CostGetDataRequest']


class CostGetDataRequest(BaseModel):
    options: dict
    secret_data: dict
    schema_name: Union[str, None] = Field(None, alias='schema')
    task_options: dict
    domain_id: str



from spaceone.core.pygrpc import BaseAPI
from spaceone.api.cost_analysis.plugin import data_source_pb2, data_source_pb2_grpc
from spaceone.cost_analysis.plugin.data_source.service.data_source_service import DataSourceService


class DataSource(BaseAPI, data_source_pb2_grpc.DataSourceServicer):

    pb2 = data_source_pb2
    pb2_grpc = data_source_pb2_grpc

    def init(self, request, context):
        params, metadata = self.parse_request(request, context)
        data_source_svc = DataSourceService(metadata)
        response: dict = data_source_svc.init(params)
        return self.dict_to_message(response)

    def verify(self, request, context):
        params, metadata = self.parse_request(request, context)
        data_source_svc = DataSourceService(metadata)
        data_source_svc.verify(params)
        return self.empty()



import logging

from spaceone.core.service import *
from spaceone.cost_analysis.error import *
from spaceone.cost_analysis.manager.budget_usage_manager import BudgetUsageManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class BudgetUsageService(BaseService):
    resource = "BudgetUsage"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.budget_usage_mgr: BudgetUsageManager = self.locator.get_manager(
            "BudgetUsageManager"
        )

    @transaction(
        permission="cost-analysis:BudgetUsage.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @check_required(["domain_id"])
    @append_query_filter(
        [
            "budget_id",
            "data_source_id",
            "name",
            "date",
            "user_projects",
            "project_id",
            "workspace_id",
            "domain_id",
        ]
    )
    @append_keyword_filter(["budget_id", "name"])
    @set_query_page_limit(1000)
    def list(self, params):
        """List budget_usages

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)',
                'budget_id': 'str',
                'data_source_id': 'str',
                'name': 'str',
                'date': 'str',
                'project_id': 'str',
                'user_projects': 'list',
                'workspace_id': str,                                # injected from auth (optional)
                'domain_id': 'str',                                 # injected from auth
            }

        Returns:
            budget_usage_vos (object)
            total_count
        """

        query = params.get("query", {})
        return self.budget_usage_mgr.list_budget_usages(query)

    @transaction(
        permission="cost-analysis:BudgetUsage.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @check_required(["query", "domain_id"])
    @append_query_filter(["budget_id", "data_source_id", "workspace_id", "domain_id"])
    @append_keyword_filter(["budget_id", "name"])
    @set_query_page_limit(1000)
    def stat(self, params):
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                "budget_id": "str",
                'data_source_id': 'str',
                'workspace_id': 'str',                                # injected from auth (optional)
                'domain_id': 'str'                                    # injected from auth
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = self._set_user_project_or_project_group_filter(params)
        return self.budget_usage_mgr.stat_budget_usages(query)

    @transaction(
        permission="cost-analysis:BudgetUsage.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @check_required(["query", "query.fields", "domain_id"])
    @append_query_filter(
        ["data_source_id", "budget_id", "user_projects", "workspace_id", "domain_id"]
    )
    @append_keyword_filter(["budget_id", "name"])
    @set_query_page_limit(1000)
    def analyze(self, params):
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v2.TimeSeriesAnalyzeQuery)',
                'budget_id': 'str',
                'data_source_id': 'str',
                'user_projects': 'list',                                        # injected from auth
                'workspace_id': 'str',                                          # injected from auth (optional)
                'domain_id': 'str'                                              # injected from auth
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = self._set_user_project_or_project_group_filter(params)
        self._check_granularity(query.get("granularity"))

        return self.budget_usage_mgr.analyze_budget_usages(query)

    @staticmethod
    def _check_granularity(granularity):
        if granularity and granularity != "MONTHLY":
            raise ERROR_INVALID_PARAMETER(
                key="query.granularity", reason="Granularity is only MONTHLY."
            )

    @staticmethod
    def _set_user_project_or_project_group_filter(params):
        query = params.get("query", {})
        query["filter"] = query.get("filter", [])

        if "user_projects" in params:
            user_projects = params["user_projects"] + [None]
            query["filter"].append(
                {"k": "user_projects", "v": user_projects, "o": "in"}
            )

        return query



from spaceone.cost_analysis.plugin.data_source.model.data_source_request import *
from spaceone.cost_analysis.plugin.data_source.model.data_source_response import *
from spaceone.cost_analysis.plugin.data_source.model.job_request import *
from spaceone.cost_analysis.plugin.data_source.model.job_response import *
from spaceone.cost_analysis.plugin.data_source.model.cost_request import *
from spaceone.cost_analysis.plugin.data_source.model.cost_response import *


__all__ = ['DataSourceInitRequest', 'DataSourceVerifyRequest', 'PluginResponse', 'JobGetTaskRequest', 'TasksResponse',
           'CostGetDataRequest', 'CostsResponse']



from typing import List, Union
from pydantic import BaseModel

__all__ = ['CostsResponse']


class Cost(BaseModel):
    cost: float
    usage_quantity: Union[float, None] = None
    usage_unit: Union[str, None] = None
    provider: Union[str, None] = None
    region_code: Union[str, None] = None
    product: Union[str, None] = None
    usage_type: Union[str, None] = None
    resource: Union[str, None] = None
    tags: dict = {}
    additional_info: dict = {}
    data: dict = {}
    billed_date: str


class CostsResponse(BaseModel):
    results: List[Cost]



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class ProviderFilter(EmbeddedDocument):
    state = StringField(
        max_length=20, default="ENABLED", choices=("ENABLED", "DISABLED")
    )
    providers = ListField(StringField(), default=[])

    def to_dict(self):
        return dict(self.to_mongo())


class BudgetUsage(MongoModel):
    budget_id = StringField(max_length=40, required=True)
    name = StringField(max_length=255, default="")
    date = StringField(max_length=7, required=True)
    cost = FloatField(required=True)
    limit = FloatField(required=True)
    currency = StringField(default=None, null=True)
    provider_filter = EmbeddedDocumentField(ProviderFilter, required=True)
    budget = ReferenceField("Budget", reverse_delete_rule=CASCADE)
    resource_group = StringField(max_length=40, choices=["WORKSPACE", "PROJECT"])
    project_id = StringField(max_length=40, default=None, null=True)
    data_source_id = StringField(max_length=40)
    workspace_id = StringField(max_length=40, default=None, null=True)
    domain_id = StringField(max_length=40)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        "updatable_fields": ["name", "cost", "limit"],
        "minimal_fields": ["budget_id", "name", "date", "usd_cost", "limit"],
        "change_query_keys": {"user_projects": "project_id"},
        "ordering": ["budget_id", "date"],
        "indexes": [
            "budget_id",
            "name",
            "date",
            "resource_group",
            "data_source_id",
            "project_id",
            "workspace_id",
            "domain_id",
        ],
    }



import logging
import fnmatch

from spaceone.core.service import *
from spaceone.cost_analysis.error import *
from spaceone.cost_analysis.manager.data_source_rule_manager import (
    DataSourceRuleManager,
)
from spaceone.cost_analysis.manager.data_source_manager import DataSourceManager
from spaceone.cost_analysis.manager.identity_manager import IdentityManager
from spaceone.cost_analysis.model.data_source_rule_model import DataSourceRule

_LOGGER = logging.getLogger(__name__)

_SUPPORTED_CONDITION_KEYS = [
    "provider",
    "region_code",
    "product",
    "account",
    "usage_type",
    "resource_group",
    "resource",
    "tags.<key>",
    "additional_info.<key>",
]
_SUPPORTED_CONDITION_OPERATORS = ["eq", "contain", "not", "not_contain"]


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class DataSourceRuleService(BaseService):
    resource = "DataSourceRule"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_source_rule_mgr: DataSourceRuleManager = self.locator.get_manager(
            "DataSourceRuleManager"
        )

    @transaction(
        permission="cost-analysis:DataSourceRule.write",
        role_types=["DOMAIN_OWNER", "WORKSPACE_OWNER"],
    )
    def create(self, params):
        """Create data source rule

        Args:
            params (dict): {
                'data_source_id': 'str',
                'name': 'str',
                'conditions': 'list',
                'conditions_policy': 'str',
                'actions': 'dict',
                'options': 'dict',
                'tags': 'dict',
                'resource_group: 'str',
                'workspace_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            data_source_rule_vo (object)
        """
        return self.create_data_source_rule(params)

    @check_required(
        [
            "data_source_id",
            "conditions_policy",
            "actions",
            "resource_group",
            "domain_id",
        ]
    )
    @change_date_value(["start", "end"])
    def create_data_source_rule(self, params):
        domain_id: str = params["domain_id"]
        data_source_id: str = params["data_source_id"]
        conditions: list = params.get("conditions", [])
        conditions_policy: str = params["conditions_policy"]
        actions: dict = params["actions"]
        rule_type: str = params.get("rule_type", "CUSTOM")

        identity_mgr: IdentityManager = self.locator.get_manager("IdentityManager")

        if params["resource_group"] == "WORKSPACE":
            identity_mgr.check_workspace(params["workspace_id"], domain_id)
        else:
            params["workspace_id"] = "*"

        if conditions_policy == "ALWAYS":
            params["conditions"] = []
        else:
            if len(conditions) == 0:
                raise ERROR_REQUIRED_PARAMETER(key="conditions")
            else:
                self._check_conditions(conditions)

        self._check_actions(actions, domain_id)

        data_source_mgr: DataSourceManager = self.locator.get_manager(
            "DataSourceManager"
        )
        data_source_vo = data_source_mgr.get_data_source(data_source_id, domain_id)

        params["data_source"] = data_source_vo
        params["order"] = (
            self._get_highest_order(data_source_id, rule_type, domain_id) + 1
        )

        return self.data_source_rule_mgr.create_data_source_rule(params)

    @transaction(
        permission="cost-analysis:DataSourceRule.write",
        role_types=["DOMAIN_OWNER", "WORKSPACE_OWNER"],
    )
    @check_required(["data_source_rule_id", "domain_id"])
    @change_date_value(["end"])
    def update(self, params):
        """Update data source rule

        Args:
            params (dict): {
                'data_source_rule_id': 'str',
                'name': 'str',
                'conditions': 'list',
                'conditions_policy': 'list',
                'actions': 'dict',
                'options': 'dict'
                'tags': 'dict'
                'workspace_id':                 # injected from auth (optional)
                'domain_id': 'str'              # injected from auth
            }

        Returns:
            data_source_rule_vo (object)
        """

        data_source_rule_id = params["data_source_rule_id"]
        domain_id = params["domain_id"]
        workspace_id = params.get("workspace_id")
        conditions_policy = params.get("conditions_policy")
        conditions = params.get("conditions", [])

        data_source_rule_vo = self.data_source_rule_mgr.get_data_source_rule(
            data_source_rule_id, domain_id, workspace_id
        )

        if data_source_rule_vo.rule_type == "MANAGED":
            raise ERROR_NOT_ALLOWED_TO_UPDATE_RULE()

        if conditions_policy:
            if conditions_policy == "ALWAYS":
                params["conditions"] = []
            else:
                if len(conditions) == 0:
                    raise ERROR_REQUIRED_PARAMETER(key="conditions")
                else:
                    self._check_conditions(conditions)

        if "actions" in params:
            self._check_actions(params["actions"], domain_id)

        return self.data_source_rule_mgr.update_data_source_rule_by_vo(
            params, data_source_rule_vo
        )

    @transaction(
        permission="cost-analysis:DataSourceRule.write",
        role_types=["DOMAIN_OWNER", "WORKSPACE_OWNER"],
    )
    @check_required(["data_source_rule_id", "order", "domain_id"])
    def change_order(self, params):
        """Change data source rule's order

        Args:
            params (dict): {
                'data_source_rule_id': 'str',
                'order': 'int',
                'workspace_id': 'str',          # injected from auth (optional)
                'domain_id': 'str'              # injected from auth
            }

        Returns:
            data_source_rule_vo (object)
        """

        data_source_rule_id = params["data_source_rule_id"]
        order = params["order"]
        domain_id = params["domain_id"]
        workspace_id = params.get("workspace_id")

        self._check_order(order)

        target_data_source_rule_vo: DataSourceRule = (
            self.data_source_rule_mgr.get_data_source_rule(
                data_source_rule_id, domain_id, workspace_id
            )
        )

        if target_data_source_rule_vo.rule_type == "MANAGED":
            raise ERROR_NOT_ALLOWED_TO_CHANGE_ORDER()

        if target_data_source_rule_vo.order == order:
            return target_data_source_rule_vo

        highest_order = self._get_highest_order(
            target_data_source_rule_vo.data_source_id,
            target_data_source_rule_vo.rule_type,
            target_data_source_rule_vo.domain_id,
        )

        if order > highest_order:
            raise ERROR_INVALID_PARAMETER(
                key="order",
                reason=f"There is no data source rules greater than the {str(order)} order.",
            )

        data_source_rule_vos = self._get_all_data_source_rules(
            target_data_source_rule_vo.data_source_id,
            target_data_source_rule_vo.rule_type,
            target_data_source_rule_vo.domain_id,
            target_data_source_rule_vo.data_source_rule_id,
        )

        data_source_rule_vos.insert(order - 1, target_data_source_rule_vo)

        i = 0
        for data_source_rule_vo in data_source_rule_vos:
            if target_data_source_rule_vo != data_source_rule_vo:
                self.data_source_rule_mgr.update_data_source_rule_by_vo(
                    {"order": i + 1}, data_source_rule_vo
                )

            i += 1

        return self.data_source_rule_mgr.update_data_source_rule_by_vo(
            {"order": order}, target_data_source_rule_vo
        )

    @transaction(
        permission="cost-analysis:DataSourceRule.write",
        role_types=["DOMAIN_OWNER", "WORKSPACE_OWNER"],
    )
    @check_required(["data_source_rule_id", "domain_id"])
    def delete(self, params):
        """Delete data source rule

        Args:
            params (dict): {
                'data_source_rule_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        data_source_rule_id = params["data_source_rule_id"]
        domain_id = params["domain_id"]
        workspace_id = params.get("workspace_id")

        data_source_rule_vo: DataSourceRule = (
            self.data_source_rule_mgr.get_data_source_rule(
                data_source_rule_id, domain_id, workspace_id
            )
        )
        rule_type = data_source_rule_vo.rule_type

        if rule_type == "MANAGED":
            raise ERROR_NOT_ALLOWED_TO_DELETE_RULE()

        data_source_id = data_source_rule_vo.data_source_id
        self.data_source_rule_mgr.delete_data_source_rule_by_vo(data_source_rule_vo)

        data_source_rule_vos = self._get_all_data_source_rules(
            data_source_id, rule_type, domain_id
        )

        i = 0
        for data_source_rule_vo in data_source_rule_vos:
            self.data_source_rule_mgr.update_data_source_rule_by_vo(
                {"order": i + 1}, data_source_rule_vo
            )
            i += 1

    @transaction(
        permission="cost-analysis:DataSourceRule.read",
        role_types=["DOMAIN_OWNER", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @check_required(["data_source_rule_id", "domain_id"])
    def get(self, params):
        """Get data source rule

        Args:
            params (dict): {
                'data_source_rule_id': 'str',
                'workspace_id' : 'list',
                'domain_id': 'str',
            }

        Returns:
            data_source_rule_vo (object)
        """

        data_source_rule_id = params["data_source_rule_id"]
        domain_id = params["domain_id"]
        workspace_id = params.get("workspace_id")

        return self.data_source_rule_mgr.get_data_source_rule(
            data_source_rule_id, domain_id, workspace_id
        )

    @transaction(
        permission="cost-analysis:DataSourceRule.read",
        role_types=["DOMAIN_OWNER", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @check_required(["domain_id"])
    @append_query_filter(
        ["data_source_rule_id", "name", "data_source_id", "workspace_id", "domain_id"]
    )
    @append_keyword_filter(["data_source_rule_id", "name"])
    def list(self, params):
        """List data source rule

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'data_source_rule_id': 'str',
                'name': 'str',
                'data_source_id': 'str',
                'workspace_id': 'str',                          # injected from auth (optional)
                'domain_id': 'str',                             # injected from auth
            }

        Returns:
            data_source_rule_vos (object)
            total_count
        """

        query = params.get("query", {})
        return self.data_source_rule_mgr.list_data_source_rules(query)

    @transaction(
        permission="cost-analysis:DataSourceRule.read",
        role_types=["DOMAIN_OWNER", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @check_required(["query", "domain_id"])
    @append_query_filter(["workspace_id", "domain_id"])
    @append_keyword_filter(["data_source_rule_id", "name"])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
                'workspace_id': 'str',                                  # injected from auth (optional)
                'domain_id': 'str',                                     # injected from auth
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get("query", {})
        return self.data_source_rule_mgr.stat_data_source_rules(query)

    @staticmethod
    def _check_conditions(conditions):
        for condition in conditions:
            key = condition.get("key")
            value = condition.get("value")
            operator = condition.get("operator")

            if not (key and value and operator):
                raise ERROR_INVALID_PARAMETER(
                    key="conditions",
                    reason="Condition should have key, value and operator.",
                )

            if key not in _SUPPORTED_CONDITION_KEYS:
                if not (
                    fnmatch.fnmatch(key, "additional_info.*")
                    or fnmatch.fnmatch(key, "tags.*")
                ):
                    raise ERROR_INVALID_PARAMETER(
                        key="conditions.key",
                        reason=f"Unsupported key. "
                        f'({" | ".join(_SUPPORTED_CONDITION_KEYS)})',
                    )
            if operator not in _SUPPORTED_CONDITION_OPERATORS:
                raise ERROR_INVALID_PARAMETER(
                    key="conditions.operator",
                    reason=f"Unsupported operator. "
                    f'({" | ".join(_SUPPORTED_CONDITION_OPERATORS)})',
                )

    def _check_actions(self, actions: dict, domain_id: str) -> None:
        if project_id := actions.get("change_project"):
            identity_mgr: IdentityManager = self.locator.get_manager("IdentityManager")
            identity_mgr.get_project(project_id, domain_id)

        if match_project := actions.get("match_project"):
            if "source" not in match_project:
                raise ERROR_REQUIRED_PARAMETER(key="actions.match_project.source")

        if match_service_account := actions.get("match_service_account"):
            if "source" not in match_service_account:
                raise ERROR_REQUIRED_PARAMETER(
                    key="actions.match_service_account.source"
                )

    def _get_highest_order(self, data_source_id, rule_type, domain_id):
        data_source_rule_vos = self.data_source_rule_mgr.filter_data_source_rules(
            data_source_id=data_source_id, rule_type=rule_type, domain_id=domain_id
        )

        return data_source_rule_vos.count()

    @staticmethod
    def _check_order(order):
        if order <= 0:
            raise ERROR_INVALID_PARAMETER(
                key="order", reason="The order must be greater than 0."
            )

    def _get_all_data_source_rules(
        self, data_source_id, rule_type, domain_id, exclude_data_source_rule_id=None
    ):
        query = {
            "filter": [
                {"k": "domain_id", "v": domain_id, "o": "eq"},
                {"k": "data_source_id", "v": data_source_id, "o": "eq"},
                {"k": "rule_type", "v": rule_type, "o": "eq"},
            ],
            "sort": [{"key": "order"}],
        }

        if exclude_data_source_rule_id is not None:
            query["filter"].append(
                {
                    "k": "data_source_rule_id",
                    "v": exclude_data_source_rule_id,
                    "o": "not",
                }
            )

        (
            data_source_rule_vos,
            total_count,
        ) = self.data_source_rule_mgr.list_data_source_rules(query)
        return list(data_source_rule_vos)



from google.protobuf.empty_pb2 import Empty
from spaceone.core.pygrpc.message_type import *

__all__ = ['EmptyInfo', 'StatisticsInfo', 'AnalyzeInfo']


def EmptyInfo():
    return Empty()


def StatisticsInfo(result):
    return change_struct_type(result)


def AnalyzeInfo(result):
    return change_struct_type(result)



import functools
from typing import List
from spaceone.api.cost_analysis.v1 import data_source_rule_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils

from spaceone.cost_analysis.model.data_source_rule_model import (
    DataSourceRule,
    DataSourceRuleCondition,
    DataSourceRuleOptions,
)

__all__ = ["DataSourceRuleInfo", "DataSourceRulesInfo"]


def DataSourceRuleConditionsInfo(condition_vos: List[DataSourceRuleCondition]):
    if condition_vos is None:
        condition_vos = []

    conditions_info = []

    for vo in condition_vos:
        info = {"key": vo.key, "value": vo.value, "operator": vo.operator}

        conditions_info.append(data_source_rule_pb2.DataSourceRuleCondition(**info))

    return conditions_info


def DataSourceRuleActionMatchRuleInfo(match_rule_data):
    if match_rule_data is None:
        return None

    info = {
        "source": match_rule_data.get("source"),
        "target": match_rule_data.get("target"),
    }

    return data_source_rule_pb2.MatchRule(**info)


def DataSourceRuleActionsInfo(actions_data):
    if actions_data is None:
        return None
    else:
        info = {}

        for key, value in actions_data.items():
            if key in ["match_project", "match_service_account"]:
                info[key] = DataSourceRuleActionMatchRuleInfo(value)
            elif key == "add_additional_info":
                info[key] = change_struct_type(value)
            else:
                info[key] = value

        return data_source_rule_pb2.DataSourceRuleActions(**info)


def DataSourceRuleOptionsInfo(vo: DataSourceRuleOptions):
    if vo is None:
        return None
    else:
        info = {"stop_processing": vo.stop_processing}

        return data_source_rule_pb2.DataSourceRuleOptions(**info)


def DataSourceRuleInfo(data_source_rule_vo: DataSourceRule, minimal=False):
    info = {
        "data_source_rule_id": data_source_rule_vo.data_source_rule_id,
        "name": data_source_rule_vo.name,
        "order": data_source_rule_vo.order,
        "rule_type": data_source_rule_vo.rule_type,
        "resource_group": data_source_rule_vo.resource_group,
        "data_source_id": data_source_rule_vo.data_source_id,
        "workspace_id": data_source_rule_vo.workspace_id,
    }

    if not minimal:
        info.update(
            {
                "conditions": DataSourceRuleConditionsInfo(
                    data_source_rule_vo.conditions
                ),
                "conditions_policy": data_source_rule_vo.conditions_policy,
                "actions": DataSourceRuleActionsInfo(data_source_rule_vo.actions),
                "options": DataSourceRuleOptionsInfo(data_source_rule_vo.options),
                "tags": change_struct_type(data_source_rule_vo.tags),
                "domain_id": data_source_rule_vo.domain_id,
                "created_at": utils.datetime_to_iso8601(data_source_rule_vo.created_at),
            }
        )

    return data_source_rule_pb2.DataSourceRuleInfo(**info)


def DataSourceRulesInfo(data_source_rule_vos, total_count, **kwargs):
    return data_source_rule_pb2.DataSourceRulesInfo(
        results=list(
            map(functools.partial(DataSourceRuleInfo, **kwargs), data_source_rule_vos)
        ),
        total_count=total_count,
    )



from spaceone.api.cost_analysis.v1 import job_pb2, job_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class Job(BaseAPI, job_pb2_grpc.JobServicer):

    pb2 = job_pb2
    pb2_grpc = job_pb2_grpc

    def cancel(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('JobService', metadata) as job_service:
            return self.locator.get_info('JobInfo', job_service.cancel(params))

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('JobService', metadata) as job_service:
            return self.locator.get_info('JobInfo', job_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('JobService', metadata) as job_service:
            job_vos, total_count = job_service.list(params)
            return self.locator.get_info('JobsInfo',
                                         job_vos,
                                         total_count,
                                         minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('JobService', metadata) as job_service:
            return self.locator.get_info('StatisticsInfo', job_service.stat(params))



from spaceone.core.pygrpc.server import GRPCServer
from spaceone.cost_analysis.plugin.data_source.interface.grpc.data_source import DataSource
from spaceone.cost_analysis.plugin.data_source.interface.grpc.job import Job
from spaceone.cost_analysis.plugin.data_source.interface.grpc.cost import Cost


_all_ = ['app']

app = GRPCServer()
app.add_service(DataSource)
app.add_service(Job)
app.add_service(Cost)



from spaceone.core.pygrpc.server import GRPCServer
from spaceone.core.plugin.server import PluginServer
from spaceone.cost_analysis.plugin.data_source.interface.grpc import app
from spaceone.cost_analysis.plugin.data_source.service.data_source_service import DataSourceService
from spaceone.cost_analysis.plugin.data_source.service.job_service import JobService
from spaceone.cost_analysis.plugin.data_source.service.cost_service import CostService

__all__ = ['DataSourcePluginServer']


class DataSourcePluginServer(PluginServer):

    _grpc_app: GRPCServer = app
    _global_conf_path: str = 'spaceone.cost_analysis.plugin.data_source.conf.global_conf:global_conf'
    _plugin_methods = {
        'DataSource': {
            'service': DataSourceService,
            'methods': ['init', 'verify']
        },
        'Job': {
            'service': JobService,
            'methods': ['get_tasks']
        },
        'Cost': {
            'service': CostService,
            'methods': ['get_data']
        }
    }



from spaceone.api.cost_analysis.v1 import data_source_rule_pb2, data_source_rule_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class DataSourceRule(BaseAPI, data_source_rule_pb2_grpc.DataSourceRuleServicer):

    pb2 = data_source_rule_pb2
    pb2_grpc = data_source_rule_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceRuleService', metadata) as data_source_rule_service:
            return self.locator.get_info('DataSourceRuleInfo', data_source_rule_service.create(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceRuleService', metadata) as data_source_rule_service:
            return self.locator.get_info('DataSourceRuleInfo', data_source_rule_service.update(params))

    def change_order(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceRuleService', metadata) as data_source_rule_service:
            return self.locator.get_info('DataSourceRuleInfo', data_source_rule_service.change_order(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceRuleService', metadata) as data_source_rule_service:
            data_source_rule_service.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceRuleService', metadata) as data_source_rule_service:
            return self.locator.get_info('DataSourceRuleInfo', data_source_rule_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceRuleService', metadata) as data_source_rule_service:
            data_source_rule_vos, total_count = data_source_rule_service.list(params)
            return self.locator.get_info('DataSourceRulesInfo',
                                         data_source_rule_vos,
                                         total_count,
                                         minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceRuleService', metadata) as data_source_rule_service:
            return self.locator.get_info('StatisticsInfo', data_source_rule_service.stat(params))



import logging
from spaceone.core.service import *
from spaceone.cost_analysis.error import *
from spaceone.cost_analysis.service.job_service import JobService
from spaceone.cost_analysis.manager.repository_manager import RepositoryManager
from spaceone.cost_analysis.manager.secret_manager import SecretManager
from spaceone.cost_analysis.manager.data_source_plugin_manager import (
    DataSourcePluginManager,
)
from spaceone.cost_analysis.manager.budget_usage_manager import BudgetUsageManager
from spaceone.cost_analysis.manager.cost_manager import CostManager
from spaceone.cost_analysis.model.data_source_model import DataSource
from spaceone.cost_analysis.manager.data_source_manager import DataSourceManager
from spaceone.cost_analysis.manager.job_manager import JobManager
from spaceone.cost_analysis.manager.identity_manager import IdentityManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class DataSourceService(BaseService):
    resource = "DataSource"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_source_mgr = DataSourceManager()
        self.ds_plugin_mgr = DataSourcePluginManager()
        self.cost_mgr = CostManager()
        self.budget_usage_mgr = BudgetUsageManager()
        self.job_mgr = JobManager()
        self.identity_mgr = IdentityManager()

    @transaction(
        permission="cost-analysis:DataSource.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(["name", "data_source_type", "domain_id"])
    def register(self, params):
        """Register data source

        Args:
            params (dict): {
                'name': 'str',              # required
                'data_source_type': 'str',  # required
                'provider': 'str',
                'secret_type': 'str',
                'secret_filter': 'dict',
                'template': 'dict',
                'plugin_info': 'dict',
                'tags': 'dict',
                'resource_group': 'str      # required
                'workspace_id': 'str'
                'domain_id': 'str'          # injected from auth
            }

        Returns:
            data_source_vo (object)
        """

        domain_id = params["domain_id"]
        data_source_type = params["data_source_type"]
        resource_group = params["resource_group"]

        # Check permission by resource group
        if resource_group == "WORKSPACE":
            self.identity_mgr.check_workspace(params["workspace_id"], domain_id)
        else:
            params["workspace_id"] = "*"

        if data_source_type == "EXTERNAL":
            params["template"] = None

            plugin_info = params.get("plugin_info", {})
            secret_type = params.get("secret_type", "MANUAL")

            if secret_type == "USE_SERVICE_ACCOUNT_SECRET" and "provider" not in params:
                raise ERROR_REQUIRED_PARAMETER(key="provider")

            self._validate_plugin_info(plugin_info, secret_type)
            self._check_plugin(plugin_info["plugin_id"])

            if "secret_filter" in params:
                self.validate_secret_filter(
                    params["secret_filter"], params["domain_id"]
                )

            # Update metadata
            (
                endpoint,
                updated_version,
            ) = self.ds_plugin_mgr.get_data_source_plugin_endpoint(
                plugin_info, domain_id
            )
            if updated_version:
                params["plugin_info"]["version"] = updated_version

            options = params["plugin_info"].get("options", {})

            plugin_metadata = self._init_plugin(endpoint, options, domain_id)

            params["plugin_info"]["metadata"] = plugin_metadata

            secret_data = plugin_info.get("secret_data")
            if secret_type == "MANUAL" and secret_data:
                self._verify_plugin(endpoint, plugin_info, domain_id)

                secret_mgr: SecretManager = self.locator.get_manager("SecretManager")
                secret_id = secret_mgr.create_secret(
                    secret_data,
                    resource_group,
                    plugin_info.get("schema_id"),
                    params["workspace_id"],
                )

                params["plugin_info"]["secret_id"] = secret_id
                del params["plugin_info"]["secret_data"]

        else:
            params["plugin_info"] = None
            params["secret_type"] = None
            params["secret_filter"] = None

            if template := params.get("template"):
                # Check Template
                pass
            else:
                raise ERROR_REQUIRED_PARAMETER(key="template")

        data_source_vo: DataSource = self.data_source_mgr.register_data_source(params)

        if data_source_type == "EXTERNAL":
            resource_group = data_source_vo.resource_group
            workspace_id = data_source_vo.workspace_id
            data_source_id = data_source_vo.data_source_id
            metadata = data_source_vo.plugin_info.metadata

            self.ds_plugin_mgr.create_data_source_rules_by_metadata(
                metadata, resource_group, data_source_id, workspace_id, domain_id
            )

        return data_source_vo

    @transaction(
        permission="cost-analysis:DataSource.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(["data_source_id", "domain_id"])
    def update(self, params):
        """Update data source

        Args:
            params (dict): {
                'data_source_id': 'str',    # required
                'name': 'str',
                'secret_filter': 'dict',
                'template': 'dict',
                'tags': 'dict'
                'domain_id': 'str'          # injected from auth
            }

        Returns:
            data_source_vo (object)
        """
        data_source_id = params["data_source_id"]
        domain_id = params["domain_id"]
        data_source_vo: DataSource = self.data_source_mgr.get_data_source(
            data_source_id, domain_id
        )

        if "secret_filter" in params:
            if data_source_vo.secret_type == "USE_SERVICE_ACCOUNT_SECRET":
                self.validate_secret_filter(
                    params["secret_filter"], params["domain_id"]
                )
            else:
                raise ERROR_NOT_ALLOW_SECRET_FILTER(data_source_id=data_source_id)

        if "template" in params:
            if data_source_vo.data_source_type == "LOCAL":
                # Check Template
                pass
            else:
                raise ERROR_NOT_ALLOW_PLUGIN_SETTINGS(data_source_id=data_source_id)

        return self.data_source_mgr.update_data_source_by_vo(params, data_source_vo)

    @transaction(
        permission="cost-analysis:DataSource.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(["data_source_id", "domain_id"])
    def verify_plugin(self, params):
        """Verify data source plugin

        Args:
            params (dict): {
                'data_source_id': 'str',    # required
                'domain_id': 'str'          # injected from auth
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params["data_source_id"]
        domain_id = params["domain_id"]
        data_source_vo: DataSource = self.data_source_mgr.get_data_source(
            data_source_id, domain_id
        )

        if data_source_vo.data_source_type == "LOCAL":
            raise ERROR_NOT_ALLOW_PLUGIN_SETTINGS(data_source_id=data_source_id)

        endpoint = self.ds_plugin_mgr.get_data_source_plugin_endpoint_by_vo(
            data_source_vo
        )
        plugin_info = data_source_vo.plugin_info.to_dict()

        self._verify_plugin(endpoint, plugin_info, domain_id)

    @transaction(
        permission="cost-analysis:DataSource.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(["data_source_id", "domain_id"])
    def update_plugin(self, params):
        """Update data source plugin

        Args:
            params (dict): {
                'data_source_id': 'str',  # required
                'version': 'str',
                'options': 'dict',
                'upgrade_mode': 'str',
                'domain_id': 'str'       # injected from auth
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params["data_source_id"]
        domain_id = params["domain_id"]
        version = params.get("version")
        options = params.get("options")
        upgrade_mode = params.get("upgrade_mode")

        data_source_vo = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        if data_source_vo.data_source_type == "LOCAL":
            raise ERROR_NOT_ALLOW_PLUGIN_SETTINGS(data_source_id=data_source_id)

        plugin_info = data_source_vo.plugin_info.to_dict()

        if version:
            plugin_info["version"] = version

        if isinstance(options, dict):
            plugin_info["options"] = options

        if upgrade_mode:
            plugin_info["upgrade_mode"] = upgrade_mode

        endpoint, updated_version = self.ds_plugin_mgr.get_data_source_plugin_endpoint(
            plugin_info, domain_id
        )
        if updated_version:
            plugin_info["version"] = updated_version

        options = plugin_info.get("options", {})
        plugin_metadata = self._init_plugin(endpoint, options, domain_id)
        plugin_info["metadata"] = plugin_metadata

        params = {"plugin_info": plugin_info}

        data_source_vo = self.data_source_mgr.update_data_source_by_vo(
            params, data_source_vo
        )

        resource_group = data_source_vo.resource_group
        workspace_id = data_source_vo.workspace_id
        self.ds_plugin_mgr.delete_data_source_rules(data_source_id, domain_id)
        self.ds_plugin_mgr.create_data_source_rules_by_metadata(
            plugin_metadata, resource_group, data_source_id, workspace_id, domain_id
        )

        return data_source_vo

    @transaction(
        permission="cost-analysis:DataSource.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(["data_source_id", "domain_id"])
    def enable(self, params):
        """Enable data source

        Args:
            params (dict): {
                'data_source_id': 'str',  # required
                'domain_id': 'str'        # injected from auth
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params["data_source_id"]
        domain_id = params["domain_id"]
        data_source_vo: DataSource = self.data_source_mgr.get_data_source(
            data_source_id, domain_id
        )

        return self.data_source_mgr.update_data_source_by_vo(
            {"state": "ENABLED"}, data_source_vo
        )

    @transaction(
        permission="cost-analysis:DataSource.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(["data_source_id", "domain_id"])
    def disable(self, params):
        """Disable data source

        Args:
            params (dict): {
                'data_source_id': 'str',    # required
                'domain_id': 'str'          # injected from auth
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params["data_source_id"]
        domain_id = params["domain_id"]
        data_source_vo: DataSource = self.data_source_mgr.get_data_source(
            data_source_id, domain_id
        )

        return self.data_source_mgr.update_data_source_by_vo(
            {"state": "DISABLED"}, data_source_vo
        )

    @transaction(
        permission="cost-analysis:DataSource.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(["data_source_id", "domain_id"])
    def deregister(self, params):
        """Deregister data source

        Args:
            params (dict): {
                'data_source_id': 'str',        # required
                'cascade_delete_cost: 'bool',
                'workspace_id: 'str',           # injected from auth (optional)
                'domain_id': 'str'              # injected from auth
            }

        Returns:
            None
        """

        data_source_id = params["data_source_id"]
        cascade_delete_cost = params.get("cascade_delete_cost", True)
        workspace_id = params.get("workspace_id")
        domain_id = params["domain_id"]

        data_source_vo: DataSource = self.data_source_mgr.get_data_source(
            data_source_id, domain_id, workspace_id
        )

        workspace_id = data_source_vo.workspace_id

        if cascade_delete_cost:
            self.cost_mgr.delete_cost_with_datasource(domain_id, data_source_id)
            self.budget_usage_mgr.update_budget_usage(
                domain_id, data_source_id
            )
            self.cost_mgr.remove_stat_cache(domain_id, data_source_id)
            self.job_mgr.preload_cost_stat_queries(domain_id, data_source_id)

        if data_source_vo.plugin_info:
            secret_id = data_source_vo.plugin_info.secret_id

            if secret_id:
                secret_mgr: SecretManager = self.locator.get_manager("SecretManager")
                secret_mgr.delete_secret(secret_id)

        self.data_source_mgr.deregister_data_source_by_vo(data_source_vo)

    @transaction(
        permission="cost-analysis:DataSource.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER"],
    )
    @check_required(["data_source_id", "domain_id"])
    def sync(self, params):
        """Sync data with data source

        Args:
            params (dict): {
                'data_source_id': 'str',        # required
                'start': 'datetime',
                'no_preload_cache': 'bool',
                'workspace_id: 'str',           # injected from auth (optional)
                'domain_id': 'str'              # injected from auth
            }

        Returns:
            None
        """
        job_service: JobService = self.locator.get_service("JobService")

        data_source_id = params["data_source_id"]
        workspace_id = params.get("workspace_id")
        domain_id = params["domain_id"]
        job_options = {
            "no_preload_cache": params.get("no_preload_cache", False),
            "start": params.get("start"),
            "sync_mode": "MANUAL",
        }

        data_source_vo: DataSource = self.data_source_mgr.get_data_source(
            data_source_id, domain_id, workspace_id
        )

        if data_source_vo.state == "DISABLED":
            raise ERROR_DATA_SOURCE_STATE(data_source_id=data_source_id)

        if data_source_vo.data_source_type == "LOCAL":
            raise ERROR_NOT_ALLOW_SYNC_COMMAND(data_source_id=data_source_id)

        return job_service.create_cost_job(data_source_vo, job_options)

    @transaction(
        permission="cost-analysis:DataSource.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @check_required(["data_source_id", "domain_id"])
    def get(self, params):
        """Get data source

        Args:
            params (dict): {
                'data_source_id': 'str',  # required
                'workspace_id': 'str'
                'domain_id': 'str',      # injected from auth
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params["data_source_id"]
        domain_id = params["domain_id"]
        workspace_id = params.get("workspace_id")

        return self.data_source_mgr.get_data_source(
            data_source_id, domain_id, workspace_id
        )

    @transaction(
        permission="cost-analysis:DataSource.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @check_required(["domain_id"])
    @append_query_filter(
        [
            "data_source_id",
            "name",
            "state",
            "data_source_type",
            "provider",
            "workspace_id",
            "domain_id",
        ]
    )
    @change_tag_filter("tags")
    @append_keyword_filter(["data_source_id", "name"])
    def list(self, params):
        """List data sources

        Args:
            params (dict): {
                'data_source_id': 'str',
                'name': 'str',
                'state': 'str',
                'data_source_type': 'str',
                'provider': 'str',
                'workspace_id': 'list,
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            data_source_vos (object)
            total_count
        """

        query = params.get("query", {})
        return self.data_source_mgr.list_data_sources(query)

    @transaction(
        permission="cost-analysis:DataSource.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @change_value_by_rule("APPEND", "workspace_id", "*")
    @check_required(["query", "domain_id"])
    @append_query_filter(["workspace_id", "domain_id"])
    @change_tag_filter("tags")
    @append_keyword_filter(["data_source_id", "name"])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get("query", {})
        return self.data_source_mgr.stat_data_sources(query)

    def validate_secret_filter(self, secret_filter, domain_id):
        if "secrets" in secret_filter:
            _query = {
                "filter": [{"k": "secret_id", "v": secret_filter["secrets"], "o": "in"}]
            }
            secret_mgr: SecretManager = self.locator.get_manager(SecretManager)
            response = secret_mgr.list_secrets(_query)
            if response.get("total_count", 0) != len(secret_filter["secrets"]):
                raise ERROR_INVALID_PARAMETER(
                    key="secret_filter.secrets", reason="Secrets not found"
                )

        if "service_accounts" in secret_filter:
            _query = {
                "filter": [
                    {
                        "k": "service_account_id",
                        "v": secret_filter["service_accounts"],
                        "o": "in",
                    }
                ]
            }
            identity_mgr: IdentityManager = self.locator.get_manager(IdentityManager)
            response = identity_mgr.list_service_accounts(_query, domain_id)
            if response.get("total_count", 0) != len(secret_filter["service_accounts"]):
                raise ERROR_INVALID_PARAMETER(
                    key="secret_filter.service_accounts",
                    reason="Service accounts not found",
                )

        if "schemas" in secret_filter:
            _query = {
                "filter": [{"k": "name", "v": secret_filter["schemas"], "o": "in"}]
            }
            repo_mgr: RepositoryManager = self.locator.get_manager(RepositoryManager)
            response = repo_mgr.list_schemas(_query, domain_id)
            if response.get("total_count", 0) != len(secret_filter["schemas"]):
                raise ERROR_INVALID_PARAMETER(
                    key="secret_filter.schema", reason="Schema not found"
                )

    def _check_plugin(self, plugin_id: str) -> None:
        repo_mgr: RepositoryManager = self.locator.get_manager("RepositoryManager")
        repo_mgr.get_plugin(plugin_id)

    def _init_plugin(self, endpoint, options, domain_id):
        self.ds_plugin_mgr.initialize(endpoint)
        return self.ds_plugin_mgr.init_plugin(options, domain_id)

    def _verify_plugin(self, endpoint, plugin_info, domain_id):
        options = plugin_info.get("options", {})
        secret_id = plugin_info.get("secret_id")
        secret_data = plugin_info.get("secret_data")
        schema = plugin_info.get("schema")

        if not secret_data:
            secret_data = self._get_secret_data(secret_id, domain_id)

        self.ds_plugin_mgr.initialize(endpoint)
        self.ds_plugin_mgr.verify_plugin(options, secret_data, schema, domain_id)

    def _get_secret_data(self, secret_id, domain_id):
        secret_mgr: SecretManager = self.locator.get_manager("SecretManager")
        if secret_id:
            secret_data = secret_mgr.get_secret_data(secret_id, domain_id)
        else:
            secret_data = {}

        return secret_data

    @staticmethod
    def _validate_plugin_info(plugin_info, secret_type):
        if "plugin_id" not in plugin_info:
            raise ERROR_REQUIRED_PARAMETER(key="plugin_info.plugin_id")

        if (
            plugin_info.get("upgrade_mode", "AUTO") == "MANUAL"
            and "version" not in plugin_info
        ):
            raise ERROR_REQUIRED_PARAMETER(key="plugin_info.version")

        if secret_type == "MANUAL" and plugin_info.get("secret_data") is None:
            raise ERROR_REQUIRED_PARAMETER(key="plugin_info.secret_data")



import logging
from typing import Generator, Union
from spaceone.core.service import BaseService, transaction, convert_model
from spaceone.cost_analysis.plugin.data_source.model import (CostGetDataRequest, CostsResponse)

_LOGGER = logging.getLogger(__name__)


class CostService(BaseService):

    @transaction
    @convert_model
    def get_data(self, params: CostGetDataRequest) -> Generator[Union[CostsResponse, dict], None, None]:
        """ Get external cost data

        Args:
            params (CostGetDataRequest): {
                'options': 'dict',          # Required
                'secret_data': 'dict',      # Required
                'schema': 'str',
                'task_options': 'dict',     # Required
                'domain_id': 'str'          # Required
            }

        Returns:
            Generator[CostsResponse, None, None]
            {
                'cost': 'float',
                'usage_quantity': 'float',
                'usage_unit': 'str',
                'provider': 'str',
                'region_code': 'str',
                'product': 'str',
                'usage_type': 'str',
                'resource': 'str',
                'tags': 'dict'
                'additional_info': 'dict'
                'data': 'dict'
                'billed_date': 'str'
            }
        """

        func = self.get_plugin_method('get_data')
        response_iterator = func(params.dict())
        for response in response_iterator:
            yield CostsResponse(**response)



import logging

from spaceone.api.cost_analysis.v1 import data_source_pb2, data_source_pb2_grpc
from spaceone.core.pygrpc import BaseAPI

_LOGGER = logging.getLogger(__name__)


class DataSource(BaseAPI, data_source_pb2_grpc.DataSourceServicer):

    pb2 = data_source_pb2
    pb2_grpc = data_source_pb2_grpc

    def register(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceService', metadata) as data_source_service:
            return self.locator.get_info('DataSourceInfo', data_source_service.register(params))

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceService', metadata) as data_source_service:
            return self.locator.get_info('DataSourceInfo', data_source_service.update(params))

    def update_plugin(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceService', metadata) as data_source_service:
            return self.locator.get_info('DataSourceInfo', data_source_service.update_plugin(params))

    def verify_plugin(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceService', metadata) as data_source_service:
            data_source_service.verify_plugin(params)
            return self.locator.get_info('EmptyInfo')

    def enable(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceService', metadata) as data_source_service:
            return self.locator.get_info('DataSourceInfo', data_source_service.enable(params))

    def disable(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceService', metadata) as data_source_service:
            return self.locator.get_info('DataSourceInfo', data_source_service.disable(params))

    def deregister(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceService', metadata) as data_source_service:
            data_source_service.deregister(params)
            return self.locator.get_info('EmptyInfo')

    def sync(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceService', metadata) as data_source_service:
            return self.locator.get_info('JobInfo', data_source_service.sync(params))

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceService', metadata) as data_source_service:
            return self.locator.get_info('DataSourceInfo', data_source_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceService', metadata) as data_source_service:
            data_source_vos, total_count = data_source_service.list(params)
            return self.locator.get_info('DataSourcesInfo',
                                         data_source_vos,
                                         total_count,
                                         minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DataSourceService', metadata) as data_source_service:
            return self.locator.get_info('StatisticsInfo', data_source_service.stat(params))



from spaceone.core.pygrpc.server import GRPCServer
from .budget import Budget
from .budget_usage import BudgetUsage
from .cost import Cost
from .cost_query_set import CostQuerySet
from .data_source import DataSource
from .data_source_rule import DataSourceRule
from .job import Job
from .job_task import JobTask


_all_ = ['app']

app = GRPCServer()
app.add_service(Budget)
app.add_service(BudgetUsage)
app.add_service(Cost)
app.add_service(CostQuerySet)
app.add_service(DataSource)
app.add_service(DataSourceRule)
app.add_service(Job)
app.add_service(JobTask)



from spaceone.core.error import *


class ERROR_NOT_ALLOWED_TO_UPDATE_RULE(ERROR_INVALID_ARGUMENT):
    _message = 'If rule_type is MANAGED, it cannot be updated.'


class ERROR_NOT_ALLOWED_TO_CHANGE_ORDER(ERROR_INVALID_ARGUMENT):
    _message = 'If rule_type is MANAGED, it cannot be changed.'


class ERROR_NOT_ALLOWED_TO_DELETE_RULE(ERROR_INVALID_ARGUMENT):
    _message = 'If rule_type is MANAGED, it cannot be deleted.'



from spaceone.api.cost_analysis.v1 import cost_pb2, cost_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class Cost(BaseAPI, cost_pb2_grpc.CostServicer):

    pb2 = cost_pb2
    pb2_grpc = cost_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostService', metadata) as cost_service:
            return self.locator.get_info('CostInfo', cost_service.create(params))

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostService', metadata) as cost_service:
            cost_service.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostService', metadata) as cost_service:
            return self.locator.get_info('CostInfo', cost_service.get(params))

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostService', metadata) as cost_service:
            cost_vos, total_count = cost_service.list(params)
            return self.locator.get_info('CostsInfo',
                                         cost_vos,
                                         total_count,
                                         minimal=self.get_minimal(params))

    def analyze(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostService', metadata) as cost_service:
            return self.locator.get_info('StatisticsInfo', cost_service.analyze(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('CostService', metadata) as cost_service:
            return self.locator.get_info('StatisticsInfo', cost_service.stat(params))



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class DataSourceRuleCondition(EmbeddedDocument):
    key = StringField(required=True)
    value = StringField(required=True)
    operator = StringField(choices=("eq", "contain", "not", "not_contain"))


class DataSourceRuleOptions(EmbeddedDocument):
    stop_processing = BooleanField(default=False)


class DataSourceRule(MongoModel):
    data_source_rule_id = StringField(max_length=40, generate_id="rule", unique=True)
    name = StringField(max_length=255, default="")
    order = IntField(required=True)
    conditions = ListField(EmbeddedDocumentField(DataSourceRuleCondition), default=[])
    conditions_policy = StringField(max_length=20, choices=("ALL", "ANY", "ALWAYS"))
    actions = DictField(default={})
    rule_type = StringField(
        max_length=255, default="CUSTOM", choices=("MANAGED", "CUSTOM")
    )
    options = EmbeddedDocumentField(
        DataSourceRuleOptions, default=DataSourceRuleOptions
    )
    tags = DictField(default={})
    data_source = ReferenceField("DataSource", reverse_delete_rule=CASCADE)
    resource_group = StringField(max_length=40, choices=["DOMAIN", "WORKSPACE"])
    data_source_id = StringField(max_length=40)
    workspace_id = StringField(max_length=40)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)

    meta = {
        "updatable_fields": [
            "name",
            "order",
            "conditions",
            "conditions_policy",
            "actions",
            "options",
            "tags",
        ],
        "minimal_fields": [
            "data_source_rule_id",
            "name",
            "order",
            "rule_type",
            "data_source_id",
        ],
        "ordering": ["order"],
        "indexes": [
            # 'data_source_rule_id',
            "order",
            "conditions_policy",
            "resource_group",
            "data_source_id",
            "workspace_id",
            "domain_id",
        ],
    }



from spaceone.cost_analysis.model.data_source_model import DataSource
from spaceone.cost_analysis.model.data_source_rule_model import DataSourceRule
from spaceone.cost_analysis.model.cost_model import Cost, MonthlyCost, CostQueryHistory
from spaceone.cost_analysis.model.budget_model import Budget
from spaceone.cost_analysis.model.budget_usage_model import BudgetUsage
from spaceone.cost_analysis.model.cost_query_set_model import CostQuerySet
from spaceone.cost_analysis.model.job_model import Job
from spaceone.cost_analysis.model.job_task_model import JobTask



from spaceone.core.error import *


class ERROR_INVALID_PLUGIN_VERSION(ERROR_INVALID_ARGUMENT):
    _message = 'Plugin version is invalid. (plugin_id = {plugin_id}, version = {version})'


class ERROR_NOT_ALLOW_PLUGIN_SETTINGS(ERROR_INVALID_ARGUMENT):
    _message = 'Local type dose not allow plugin settings. (data_source_id = {data_source_id})'


class ERROR_NOT_ALLOW_SYNC_COMMAND(ERROR_INVALID_ARGUMENT):
    _message = 'Local type dose not allow sync commands. (data_source_id = {data_source_id})'


class ERROR_DATA_SOURCE_STATE(ERROR_INVALID_ARGUMENT):
    _message = 'Data source is disabled (data_source_id = {data_source_id})'


class ERROR_NOT_ALLOW_SECRET_FILTER(ERROR_INVALID_ARGUMENT):
    _message = 'USE_SERVICE_ACCOUNT_SECRET type dose not allow secret filter. (data_source_id = {data_source_id})'

